@@ -9,6 +9,7 @@ from gi.repository import GObject, Gdk, Gtk, Gio
 from UbuntuDrivers import detect
 from aptdaemon import client
 from aptdaemon.errors import NotAuthorizedError, TransactionFailed
+from aptdaemon.gtk3widgets import AptErrorDialog, AptProgressDialog
 import re
 import urllib
 
@@ -17,8 +18,8 @@ gettext.install("mintdrivers", "/usr/share/linuxmint/locale")
 class Application():
 
   def __init__(self):
-            
-    self.builder = Gtk.Builder()    
+
+    self.builder = Gtk.Builder()
     self.builder.add_from_file("/usr/lib/linuxmint/mintDrivers/main.ui")
     self.builder.connect_signals(self)
     for o in self.builder.get_objects():
@@ -34,12 +35,35 @@ class Application():
 
     self.window_main.connect("delete_event", self.quit_application)
 
-    self.apt_cache = apt.Cache()
+    self.button_driver_revert = Gtk.Button(label=_("Re_vert"), use_underline=True)
+    self.button_driver_revert.connect("clicked", self.on_driver_changes_revert)
+    self.button_driver_apply = Gtk.Button(label=_("_Apply Changes"), use_underline=True)
+    self.button_driver_apply.connect("clicked", self.on_driver_changes_apply)
+    self.button_driver_cancel = Gtk.Button(label=_("_Cancel"), use_underline=True)
+    self.button_driver_cancel.connect("clicked", self.on_driver_changes_cancel)
+    self.button_driver_restart = Gtk.Button(label=_("_Restart..."), use_underline=True)
+    self.button_driver_restart.connect("clicked", self.on_driver_restart_clicked)
+    self.button_driver_revert.set_sensitive(False)
+    self.button_driver_revert.set_visible(True)
+    self.button_driver_apply.set_sensitive(False)
+    self.button_driver_apply.set_visible(True)
+    self.button_driver_cancel.set_visible(False)
+    self.button_driver_restart.set_visible(False)
+    self.box_driver_action.pack_end(self.button_driver_apply, False, False, 0)
+    self.box_driver_action.pack_end(self.button_driver_revert, False, False, 0)
+    self.box_driver_action.pack_end(self.button_driver_restart, False, False, 0)
+    self.box_driver_action.pack_end(self.button_driver_cancel, False, False, 0)
+
+    self.info_bar.set_no_show_all (True)
+
+    self.progress_bar = Gtk.ProgressBar()
+    self.box_driver_action.pack_end(self.progress_bar, False, False, 0)
+    self.progress_bar.set_visible(False)
+
+    self.needs_restart = False
+    self.live_mode = False
     self.apt_client = client.AptClient()
 
-    self.init_drivers()
-    self.show_drivers()  
-    
     with open('/proc/cmdline') as f:
       cmdline = f.read()
       if ((not "boot=casper" in cmdline) and (not "boot=live" in cmdline)):
@@ -49,8 +73,36 @@ class Application():
         self.button_info_bar.connect("clicked", self.check_internet_or_live_dvd)
       else:
         print ("Live mode detected")
+        self.live_mode = True
         self.info_bar.hide()
-  
+        self.update_cache()
+
+  def update_cache(self):
+    self.update_transaction = self.apt_client.update_cache()
+    self.update_transaction.connect("finished", self._on_cache_update_finished)
+    dia = AptProgressDialog(self.update_transaction)
+    dia.set_transient_for(self.window_main)
+    dia.set_modal(True)
+    dia.run(close_on_finished=True, show_error=True,
+            reply_handler=lambda: True,
+            error_handler=self.on_error,
+            )
+
+  def on_error(self, error):
+    if isinstance(error, NotAuthorizedError):
+        # Silently ignore auth failures
+        return
+    elif not isinstance(error, TransactionFailed):
+        # Catch internal errors of the client
+        error = TransactionFailed(ERROR_UNKNOWN,str(error))
+    dia = AptErrorDialog(error)
+    dia.run()
+    dia.hide()
+    self.apt_cache = apt.Cache()
+
+  def _on_cache_update_finished(self, transaction, exit_state):
+    self.show_drivers()
+
   def quit_application(self, widget=None, event=None):
     self.clean_up_media_cdrom()
     Gtk.main_quit()
@@ -72,29 +124,31 @@ class Application():
       return False
 
   def check_internet_or_live_dvd(self, widget=None):
-    
+
     try:
-      print ("Checking...")    
+      print ("Checking connectivity or live media...")
 
       # We either get drivers from the Internet of from the liveDVD
       # So check the connection to the Internet or scan for a liveDVD
-      
+
       if self.check_connectivity("http://archive.ubuntu.com"):
         # We can reach the repository, everything's fine
+        print ("Internet connection detected")
         self.info_bar.hide()
+        self.update_cache()
         return
-      
+
       # We can't reach the repository, we need the installation media
       mount_point = None
       mounted_on_media_cdrom = False
-     
+
       # Find the live media
       try:
         live_medias = subprocess.check_output("find /media | grep README.diskdefines", shell=True)
         live_medias = str(live_medias, encoding='utf8').split("\n")
         for live_media in live_medias:
           if ("README.diskdefines" in live_media):
-            mount_point = live_media.replace("/README.diskdefines", "") # This is where our live DVD is mounted 
+            mount_point = live_media.replace("/README.diskdefines", "") # This is where our live DVD is mounted
             # Add it to apt-cdrom
             os.system("sudo apt-cdrom -d \"%s\" -m add" % mount_point)
             if ("/media/cdrom" in mount_point):
@@ -104,9 +158,9 @@ class Application():
 
       if mount_point is None:
         # Not mounted..
-        self.info_bar.show()        
+        self.info_bar.show()
         return
-      
+
       if mounted_on_media_cdrom == False:
         # Mounted, but not in the right place...
         print ("Binding mount %s to /media/cdrom" % mount_point)
@@ -116,12 +170,14 @@ class Application():
 
       # It should now be mounted to /media/cdrom, else something went wrong..
       if os.path.exists("/media/cdrom/README.diskdefines"):
+        print ("Live media detected")
         self.info_bar.hide()
+        self.update_cache()
       else:
         self.info_bar.show()
     except (NotAuthorizedError, TransactionFailed) as e:
-      print("An error occured: {}".format(e))      
-      self.info_bar.show()      
+      print("An error occured: {}".format(e))
+      self.info_bar.show()
 
   def on_driver_changes_progress(self, transaction, progress):
     #print(progress)
@@ -136,6 +192,7 @@ class Application():
     self.progress_bar.set_fraction(progress / 100.0)
 
   def on_driver_changes_finish(self, transaction, exit_state):
+    self.needs_restart = True
     self.progress_bar.set_visible(False)
     self.clear_changes()
     self.apt_cache = apt.Cache()
@@ -154,6 +211,7 @@ class Application():
     self.button_driver_apply.set_visible(True)
     self.button_driver_cancel.set_visible(False)
     self.scrolled_window_drivers.set_sensitive(True)
+    self.on_error(transaction.error)
 
   def on_driver_changes_cancellable_changed(self, transaction, cancellable):
     self.button_driver_cancel.set_sensitive(cancellable)
@@ -167,7 +225,7 @@ class Application():
       if pkg.is_installed:
         removals.append(pkg.shortname)
         print ("Remove %s" % pkg.shortname)
-      else:        
+      else:
         installs.append(pkg.shortname)
         print ("Install %s" % pkg.shortname)
         try:
@@ -177,10 +235,11 @@ class Application():
               print ("Install %s" % recommend[0].name)
         except:
           print ("A problem occured, some recommended deps might not get installed")
-    
+
     try:
       self.transaction = self.apt_client.commit_packages(install=installs, remove=removals,
                                                          reinstall=[], purge=[], upgrade=[], downgrade=[])
+      self.transaction.set_allow_unauthenticated(True)
       self.transaction.connect("progress-changed", self.on_driver_changes_progress)
       self.transaction.connect("cancellable-changed", self.on_driver_changes_cancellable_changed)
       self.transaction.connect("finished", self.on_driver_changes_finish)
@@ -215,50 +274,12 @@ class Application():
 
   def on_driver_restart_clicked(self, button_restart):
     self.clean_up_media_cdrom()
-    subprocess.call(['/usr/lib/indicator-session/gtk-logout-helper', '--shutdown'])
+    subprocess.call(['reboot'])
 
   def clear_changes(self):
     self.orig_selection = {}
     self.driver_changes = []
 
-  def init_drivers(self):
-    """Additional Drivers tab"""
-
-    self.button_driver_revert = Gtk.Button(label=_("Re_vert"), use_underline=True)
-    self.button_driver_revert.connect("clicked", self.on_driver_changes_revert)
-    self.button_driver_apply = Gtk.Button(label=_("_Apply Changes"), use_underline=True)
-    self.button_driver_apply.connect("clicked", self.on_driver_changes_apply)
-    self.button_driver_cancel = Gtk.Button(label=_("_Cancel"), use_underline=True)
-    self.button_driver_cancel.connect("clicked", self.on_driver_changes_cancel)
-    self.button_driver_restart = Gtk.Button(label=_("_Restart..."), use_underline=True)
-    self.button_driver_restart.connect("clicked", self.on_driver_restart_clicked)
-    self.button_driver_revert.set_sensitive(False)
-    self.button_driver_revert.set_visible(True)
-    self.button_driver_apply.set_sensitive(False)
-    self.button_driver_apply.set_visible(True)
-    self.button_driver_cancel.set_visible(False)
-    self.button_driver_restart.set_visible(False)
-    self.box_driver_action.pack_end(self.button_driver_apply, False, False, 0)
-    self.box_driver_action.pack_end(self.button_driver_revert, False, False, 0)
-    self.box_driver_action.pack_end(self.button_driver_restart, False, False, 0)
-    self.box_driver_action.pack_end(self.button_driver_cancel, False, False, 0)
-
-    self.progress_bar = Gtk.ProgressBar()
-    self.box_driver_action.pack_end(self.progress_bar, False, False, 0)
-    self.progress_bar.set_visible(False)
-
-    self.devices = detect.system_device_drivers()
-    self.driver_changes = []
-    self.orig_selection = {}
-    # HACK: the case where the selection is actually "Do not use"; is a little
-    #       tricky to implement because you can't check for whether a package is
-    #       installed or any such thing. So let's keep a list of all the 
-    #       "Do not use" radios, set those active first, then iterate through
-    #       orig_selection when doing a Reset.
-    self.no_drv = []
-    self.nonfree_drivers = 0
-    self.ui_building = False
-
   def on_driver_selection_changed(self, button, modalias, pkg_name=None):
     if self.ui_building:
       return
@@ -274,7 +295,7 @@ class Application():
       if pkg in self.driver_changes:
         self.driver_changes.remove(pkg)
 
-      if (pkg is not None 
+      if (pkg is not None
          and modalias in self.orig_selection
          and button is not self.orig_selection[modalias]):
         self.driver_changes.append(pkg)
@@ -302,7 +323,8 @@ class Application():
         {"recommended/alternative": {pkg_name: {
                                                   'selected': True/False
                                                   'description': 'description'
-                                                  'builtin': True/False
+                                                  'builtin': True/False,
+                                                  'free': True/False
                                                 }
                                      }}
          "manually_installed": {"manual": {'selected': True, 'description': description_string}}
@@ -351,14 +373,24 @@ class Application():
       try:
         pkg = self.apt_cache[pkg_driver_name]
         installed = pkg.is_installed
-        if driver_status == 'recommended':        
-          description = ("<b>%s <small><span foreground='#58822B'>(%s)</span></small></b>\n<small><span foreground='#3c3c3c'>%s</span></small> %s\n<small><span foreground='#3c3c3c'>%s</span></small>") % (pkg.shortname, _("recommended"), _("Version"), pkg.candidate.version, pkg.candidate.summary)
-        else:
-          description = ("<b>%s</b>\n<small><span foreground='#3c3c3c'>%s</span></small> %s\n<small><span foreground='#3c3c3c'>%s</span></small>") % (pkg.shortname, _("Version"), pkg.candidate.version, pkg.candidate.summary)
+        description_line1 = "<b>%s</b>" % pkg.shortname
+        description_line2 = "<small><span foreground='#3c3c3c'>%s</span></small> %s" % (_("Version"), pkg.candidate.version)
+        description_line3 = "<small><span foreground='#3c3c3c'>%s</span></small>" % pkg.candidate.summary
+        if driver_status == 'recommended':
+          description_line1 = "%s <b><small><span foreground='#58822B'>(%s)</span></small></b>" % (description_line1, _("recommended"))
+        if current_driver['free'] and pkg.shortname != "bcmwl-kernel-source":
+          description_line1 = "%s <b><small><span foreground='#2b3882'>(%s)</span></small></b>" % (description_line1, _("open-source"))
+        if pkg.shortname == "bcmwl-kernel-source":
+          # Always recommend STA driver (B43 is always present alongside it now)
+          description_line1 = "%s <b><small><span foreground='#58822B'>(%s)</span></small></b>" % (description_line1, _("recommended"))
+        if pkg.shortname.startswith("firmware-b43"):
+          # B43 requires a connection to the Internet
+          description_line1 = "%s <b><small><span foreground='#822b32'>(%s)</span></small></b>" % (description_line1, _("requires a connection to the Internet"))
+        description = "%s\n%s\n%s" % (description_line1, description_line2, description_line3)
       except KeyError:
         print("WARNING: a driver ({}) doesn't have any available package associated: {}".format(pkg_driver_name, current_driver))
-        continue     
-          
+        continue
+
       selected = False
       if not builtin and not returned_drivers['manually_installed']:
         selected = installed
@@ -368,7 +400,8 @@ class Application():
 
       returned_drivers[driver_status].setdefault(pkg_driver_name, {'selected': selected,
                                                                    'description': description,
-                                                                   'builtin': builtin})
+                                                                   'builtin': builtin,
+                                                                   'free': current_driver['free']})
 
     # adjust making the needed addition
     if not have_builtin:
@@ -393,7 +426,7 @@ class Application():
 
     return (overall_status, icon, returned_drivers)
 
-  def get_device_icon(self, device):    
+  def get_device_icon(self, device):
     vendor = device.get('vendor', _('Unknown'))
     model =  device.get('model', _('Unknown'))
     icon = "generic"
@@ -406,9 +439,20 @@ class Application():
     elif "virtualbox" in vendor.lower():
       icon = "virtualbox"
     return ("/usr/share/linuxmint/mintDrivers/icons/%s.png" % icon)
-    
+
 
   def show_drivers(self):
+    self.apt_cache = apt.Cache()
+    self.devices = detect.system_device_drivers()
+    self.driver_changes = []
+    self.orig_selection = {}
+    # HACK: the case where the selection is actually "Do not use"; is a little
+    #       tricky to implement because you can't check for whether a package is
+    #       installed or any such thing. So let's keep a list of all the
+    #       "Do not use" radios, set those active first, then iterate through
+    #       orig_selection when doing a Reset.
+    self.no_drv = []
+    self.nonfree_drivers = 0
     self.ui_building = True
     self.dynamic_device_status = {}
     for device in sorted(self.devices.keys()):
@@ -447,7 +491,7 @@ class Application():
       option_group = None
       # define the order of introspection
       for section in ('recommended', 'alternative', 'manually_installed', 'no_driver'):
-        for driver in drivers[section]:
+        for driver in sorted(drivers[section], key=lambda x: self.sort_string(drivers[section], x)):
           radio_button = Gtk.RadioButton.new(None)
           label = Gtk.Label()
           label.set_markup(drivers[section][driver]['description'])
@@ -474,6 +518,14 @@ class Application():
     self.box_driver_detail.show_all()
     self.set_driver_action_status()
 
+  def sort_string(self, drivers, x):
+    value = drivers[x]['description']
+    try:
+      value = "%s %s" % (not drivers[x]['free'], value)
+    except:
+      pass #best effort (some driver options don't have a 'free' flag, and that's alright)
+    return value
+
   def update_label_and_icons_from_status(self):
     '''Update the current label and icon, computing the new device status'''
 
@@ -487,7 +539,7 @@ class Application():
 
   def set_driver_action_status(self):
     # Update the label in case we end up having some kind of proprietary driver in use.
-    if (os.path.exists('/var/run/reboot-required')):
+    if (not self.live_mode) and (os.path.exists('/var/run/reboot-required') or self.needs_restart):
       self.label_driver_action.set_label(_("You need to restart the computer to complete the driver changes."))
       self.button_driver_restart.set_visible(True)
       self.window_main.set_urgency_hint(True)
@@ -497,13 +549,13 @@ class Application():
     for device in self.devices:
       for pkg_name in self.devices[device]['drivers']:
         pkg = self.apt_cache[pkg_name]
-        if not self.devices[device]['drivers'][pkg_name]['free'] and pkg.is_installed:
+        if (not self.devices[device]['drivers'][pkg_name]['free'] or pkg_name == "bcmwl-kernel-source") and pkg.is_installed:
           self.nonfree_drivers = self.nonfree_drivers + 1
 
     if self.nonfree_drivers > 0:
       self.label_driver_action.set_label(gettext.ngettext (
                                                 "%(count)d proprietary driver in use.",
-                                                "%(count)d proprietary drivers in use.", 
+                                                "%(count)d proprietary drivers in use.",
                                                 self.nonfree_drivers)
                                                 % { 'count': self.nonfree_drivers})
     else:
